/**
 * Copyright (c) 2003, Brian McCallister
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name of the copyright holder nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 *THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package org.skife.intercept;

import java.lang.reflect.Method;
import java.util.Map;
import java.util.HashMap;

/**
 * Maintains lookup tables and builds Interceptor[] for methods
 */
class FastMethodInterceptMatrix
{
    /**
     * The Class being intercepted
     */
    private final Class interfaze;

    /**
     * Method -> Wrapper[]
     */
    private final Map wrappers;

    /**
     * Build an interception matrix for interfaze
     */
    public FastMethodInterceptMatrix(Class interfaze)
    {
        this.interfaze = interfaze;
        this.wrappers = new HashMap();
    }

    /**
     * Compare sig against all methods and add interceptor if it matches
     *
     * @param sig Signature to match against
     * @param interceptor <code>Interceptor</code> class to use for matches
     *        must support <code>Class.instanceOf()</code> instantiation
     */
    public synchronized void consider(Signature sig, Class interceptor)
    {
        Method[] methods = interfaze.getMethods();
        int size = methods.length;
        for (int i = 0; i < size; ++i)
        {
            if (sig.matches(methods[i]))
            {
                Wrapper[] wrappers = (Wrapper[]) this.wrappers.get(methods[i]);
                if (wrappers == null) wrappers = new Wrapper[] {};
                Wrapper[] temp = new Wrapper[wrappers.length + 1];
                System.arraycopy(wrappers, 0, temp, 0, wrappers.length);
                temp[temp.length - 1] = new FactoryWrapper(interceptor);
                this.wrappers.put(methods[i], temp);
            }
        }
    }

    /**
     * Compare sig against all methods and add interceptor if it matches
     *
     * @param sig Signature to match against
     * @param interceptor singleton Interceptor
     */
    public synchronized void consider(Signature sig, Interceptor interceptor)
    {
        Method[] methods = interfaze.getMethods();
        int size = methods.length;
        for (int i = 0; i < size; ++i)
        {
            if (sig.matches(methods[i]))
            {
                Wrapper[] wrappers = (Wrapper[]) this.wrappers.get(methods[i]);
                if (wrappers == null) wrappers = new Wrapper[] {};
                Wrapper[] temp = new Wrapper[wrappers.length + 1];
                System.arraycopy(wrappers, 0, temp, 0, wrappers.length);
                temp[temp.length - 1] = new SingletonWrapper(interceptor);
                this.wrappers.put(methods[i], temp);
            }
        }
    }

    /**
     * Obtain an array of the Interceptors for this method. This will create, or retrieve
     * singletons, as necesary.
     */
    public Interceptor[] interceptsFor(Method m) throws IllegalAccessException, InstantiationException
    {
        Wrapper[] wraps = (Wrapper[]) this.wrappers.get(m);
        if (wraps == null) return new Interceptor[] {};

        Interceptor[] intercepts = new Interceptor[wraps.length];
        int size = intercepts.length;
        for (int i = 0; i < size; ++i)
        {
            intercepts[i] = wraps[i].getIntercept();
        }
        return intercepts;
    }

    /**
     * Provide polymorphism for singleton vs instantiated interceptors.
     */
    private static interface Wrapper
    {
        Interceptor getIntercept() throws IllegalAccessException, InstantiationException;
    }

    /**
     * For singleton interceptors
     */
    private static class SingletonWrapper implements Wrapper
    {
        private final Interceptor singleton;

        SingletonWrapper(Interceptor intercept)
        {
            this.singleton = intercept;
        }

        public Interceptor getIntercept()
        {
            return this.singleton;
        }
    }

    /**
     * For per-call interceptors
     */
    private static class FactoryWrapper implements Wrapper
    {
        private final Class clazz;

        FactoryWrapper(Class clazz)
        {
            this.clazz = clazz;
        }


        public Interceptor getIntercept() throws IllegalAccessException, InstantiationException
        {
            return (Interceptor) clazz.newInstance();
        }

    }

}
