/**
 * Copyright (c) 2003, Brian McCallister
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name of the copyright holder nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 *THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package org.skife.intercept;

import net.sf.cglib.Enhancer;
import net.sf.cglib.Factory;
import net.sf.cglib.MethodInterceptor;
import net.sf.cglib.MethodProxy;

import java.util.HashMap;
import java.util.WeakHashMap;
import java.lang.reflect.Method;

/**
 * provide interception matches against arbitrary non-final objects
 */
public class InterceptionBroker
{
    /**
     * Class -> Factory
     */
    private final HashMap factories;

    /**
     * Used to build ServiceInterceptors
     */
    private final InterceptorBuilder builder;

    public InterceptionBroker()
    {
        this.factories = new HashMap();
        this.builder = new InterceptorBuilder();
    }

    /**
     * Wrap an arbitrary object in whatever interceptors match it.
     * <p>
     * It is important to note that an object can only be wrapped
     * by a single broker at a time, or to put it more specifically,
     * a broker will NOT wrap an already wrapped object -- this is
     * a meaningless triviality when attempting to wrap the same object
     * via the same broker multiple times, but it means that if an
     * InterceptedObject is passed to this method, you will get the same
     * object back with no changes.
     * <p>
     * A single object can in fact be wrapped by multiple brokers
     * concurrently, however the wrapped objects will be different, for
     * example:
     * <pre><code>
     * HashMap foo = new HashMap();
     * HashMap wrappedFooOne = (HashMap)broker1.wrap(foo);
     * HashMap wrappedFooTwo = (HashMap)broker2.wrap(foo);
     * </code></pre>
     * is perfectly okay, however calls to the two wrapped foos
     * will go through the interception stacks defined in their
     * respective brokers -- not through both.
     * <p>
     * The following code is an example of the behavior warned about:
     * <pre><code>
     * HashMap foo = new HashMap();
     * HashMap wrappedFooOne = (HashMap)broker1.wrap(foo);
     * HashMap wrappedFooTwo = (HashMap)broker2.wrap(wrappedFooOne);
     * </code></pre>
     * In this case the wrappedFooTwo will not execute
     * the interception stack defined in broker2, only the one
     * defined in broker1.
     * <p>
     * TODO Change above mentioned behavior to expected behavior
     *
     * @param obj must not be declared final
     * @return a subclass wrapping obj which will be intercepted.
     */
    public InterceptedObject wrap(final Object obj)
    {
        if (obj instanceof InterceptedObject) return (InterceptedObject) obj;
        Class clazz = obj.getClass();
        Factory factory = (Factory) this.factories.get(clazz);

        if (factory == null)
        {
            this.builder.register(clazz);
            factory = Enhancer.create(obj.getClass(), new Class[] {InterceptedObject.class }, new MethodInterceptor()
            {
                public Object intercept(Object o, Method method, Object[] objects, MethodProxy proxy)
                        throws Throwable
                {
                    return proxy.invoke(obj, objects);
                }
            });
            this.factories.put(clazz, factory);
        }
        InterceptedObject instance = (InterceptedObject) factory.newInstance(this.builder.build(clazz, obj));
        return instance;
    }

    /**
     * Used to add an interceptor where a new <code>Interceptor</code>
     * will be instantiated for each invocation.
     *
     * @param interceptor Class implementing <code>Interceptor</code>
     * @param signature Signature which will be matched for interception
     */
    public synchronized void addInterceptor(Class interceptor, Signature signature)
    {
        builder.addInterceptor(signature, interceptor);
    }

    /**
     * The <code>Interceptor</code> passed here will be used for every method
     * invocation. Make sure it is threadsafe if you are in a multithreaded environment!
     *
     * @param interceptor Singleton interceptor
     * @param signature
     */
    public synchronized void addInterceptor(Interceptor interceptor, Signature signature)
    {
        builder.addInterceptor(signature, interceptor);
    }
}
