/**
 * Copyright (c) 2003, Brian McCallister
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name of the copyright holder nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 *THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package org.skife.intercept;

import org.skife.intercept.Interceptor;

import net.sf.cglib.MethodProxy;
import net.sf.cglib.core.CodeGenerationException;

import java.lang.reflect.Method;

/**
 * Provides one way traversal across a list of interceptors. Excution of the
 * wrapped invocation should be started via InterceptorStack.yield() to yield the
 * thread of execution to the first interceptor.
 *
 * @see org.skife.intercept.Interceptor
 */
public class InterceptorStack
{
    /**
     * The intercepts in this stack
     */
    private Interceptor[] intercepts;

    /**
     * Which interceptor to execute next
     */
    private int next;

    /**
     * Command whose execute method is being intercepted
     */
    private Method method;

    /**
     * Base object whose method calls are being intercepted
     */
    private Object base;

    /**
     * Invocation arguments
     */
    private Object[] args;

    /**
     * CGLIB generated proxy (avoids reflection)
     */
    private MethodProxy proxy;


    /**
     * Create an InterceptorStack with the following interceptors on this invocation.
     * The Interceptor instances will be called in the same order in which they
     * are listed in the array.
     *
     * @param intercepts Intercepters in order they will be calles
     * @param base Base object at bottom of stack
     * @param method Method being intercepted
     * @param args invocation arguments
     */
    InterceptorStack(Interceptor[] intercepts, Object base, Method method, Object[] args, MethodProxy proxy)
    {
        this.method = method;
        this.intercepts = intercepts;
        this.args = args;
        this.base = base;
        this.proxy = proxy;
    }

    /**
     * Yield control to the next Interceptor in the stack, or to the base object.
     *
     * @return The return value from the downstream command execution
     * @throws java.lang.Exception Unhandled exception passed up the stack
     * @see org.skife.intercept.Interceptor#intercept
     */
    public Object yield() throws Throwable
    {
        if (next == intercepts.length)
        {
            return proxy.invoke(base, args);
        }
        else
        {
            return this.intercepts[next++].intercept(base, method, args, this);
        }
    }

    /**
     * Yield control to the next Interceptor in the stack, or to the base object.
     * <p>
     * This form allows overriding the complete calling environment of the downstream stack. It
     * has no effect on the upstream interceptors.
     *
     * @return The return value from the downstream invocation
     * @throws java.lang.Exception Unhandled exception passed up the stack
     * @see org.skife.intercept.Interceptor#intercept
     */
    public Object yield(Object base, Method method, Object[] args) throws Throwable
    {
        this.base = base;
        this.method = method;
        this.args = args;
        return this.yield();
    }
}
